import logging
import census

from .core import ACS5Client, SF1Client, PLClient


class Census(census.Census):
    '''
    Unified API for accessing Census data

    Attributes:

    * acs5: instance of ACS5Client
    * sf1: instance of SF1Client
    * pl: instance of PLClient
    '''
    def __init__(self, key, year=None, session=None):
        super(Census, self).__init__(key, year, session)
        self.acs5 = ACS5Client(key, year, session)
        self.sf1 = SF1Client(key, year, session)
        self.pl = PLClient(key, year, session)



GEO_URLS = {
    'tracts': {
        2000: 'https://tigerweb.geo.census.gov/arcgis/rest/services/Census2020/tigerWMS_Census2000/MapServer/8',
        2010: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_Census2010/MapServer/14',
        2011: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_Census2010/MapServer/14',
        2012: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_Census2010/MapServer/14',
        2013: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2013/MapServer/8',
        2014: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2014/MapServer/8',
        2015: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2015/MapServer/8',
        2016: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2016/MapServer/8',
        2017: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2017/MapServer/8',
        2018: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2018/MapServer/8',
        2019: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2019/MapServer/8',
        2020: 'https://tigerweb.geo.census.gov/arcgis/rest/services/Census2020/tigerWMS_Census2020/MapServer/6',
        2021: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2021/MapServer/6',
        2022: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2022/MapServer/6',
        2023: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2023/MapServer/8',
    },
    'block groups': {
        2000: 'https://tigerweb.geo.census.gov/arcgis/rest/services/Census2020/tigerWMS_Census2000/MapServer/10',
        2010: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_Census2010/MapServer/16',
        2011: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_Census2010/MapServer/16',
        2012: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_Census2010/MapServer/16',
        2013: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2013/MapServer/10',
        2014: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2014/MapServer/10',
        2015: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2015/MapServer/10',
        2016: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2016/MapServer/10',
        2017: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2017/MapServer/10',
        2018: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2017/MapServer/10',
        2019: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2018/MapServer/10',
        2020: 'https://tigerweb.geo.census.gov/arcgis/rest/services/Census2020/tigerWMS_Census2020/MapServer/8',
        2021: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2021/MapServer/8',
        2022: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2022/MapServer/8',
        2023: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2023/MapServer/10',
    },
    'blocks': {
        2000: 'https://tigerweb.geo.census.gov/arcgis/rest/services/Census2020/tigerWMS_Census2000/MapServer/12',
        2010: 'https://tigerweb.geo.census.gov/arcgis/rest/services/Census2020/tigerWMS_Census2010/MapServer/14',
        2020: 'https://tigerweb.geo.census.gov/arcgis/rest/services/Census2020/tigerWMS_Census2020/MapServer/10',
    },
    'incorporated places': {
        2000: 'https://tigerweb.geo.census.gov/arcgis/rest/services/Census2020/tigerWMS_Census2000/MapServer/26',
        2010: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_Census2010/MapServer/34',
        2011: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_Census2010/MapServer/34',
        2012: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_Census2010/MapServer/34',
        2013: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2013/MapServer/26',
        2014: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2014/MapServer/26',
        2015: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2015/MapServer/26',
        2016: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2016/MapServer/26',
        2017: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2017/MapServer/26',
        2018: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2018/MapServer/26',
        2019: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2019/MapServer/26',
        2020: 'https://tigerweb.geo.census.gov/arcgis/rest/services/Census2020/tigerWMS_Census2020/MapServer/26',
        2021: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2021/MapServer/24',
        2022: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2022/MapServer/24',
        2023: 'https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/tigerWMS_ACS2023/MapServer/28',
    },
}

LODES_VARIABLES = {
    "id": "Census Block code - 15 character nationally unique identifier",
    "c000": "Total number of jobs",
    "ca01": "Number of jobs of workers age 29 or younger",
    "ca02": "Number of jobs for workers age 30 to 54",
    "ca03": "Number of jobs for workers age 55 or older",
    "ce01": "Number of jobs with earnings $1250/month or less",
    "ce02": "Number of jobs with earnings $1251/month to $3333/month",
    "ce03": "Number of jobs with earnings greater than $3333/month",
    "cns01": "Number of jobs in NAICS sector 11 (Agriculture, Forestry, Fishing and Hunting)",
    "cns02": "Number of jobs in NAICS sector 21 (Mining, Quarrying, and Oil and Gas Extraction)",
    "cns03": "Number of jobs in NAICS sector 22 (Utilities)",
    "cns04": "Number of jobs in NAICS sector 23 (Construction)",
    "cns05": "Number of jobs in NAICS sector 31-33 (Manufacturing)",
    "cns06": "Number of jobs in NAICS sector 42 (Wholesale Trade)",
    "cns07": "Number of jobs in NAICS sector 44-45 (Retail Trade)",
    "cns08": "Number of jobs in NAICS sector 48-49 (Transportation and Warehousing)",
    "cns09": "Number of jobs in NAICS sector 51 (Information)",
    "cns10": "Number of jobs in NAICS sector 52 (Finance and Insurance)",
    "cns11": "Number of jobs in NAICS sector 53 (Real Estate and Rental and Leasing)",
    "cns12": "Number of jobs in NAICS sector 54 (Professional, Scientific, and Technical Services)",
    "cns13": "Number of jobs in NAICS sector 55 (Management of Companies and Enterprises)",
    "cns14": "Number of jobs in NAICS sector 56 (Administrative and Support and Waste Management and Remediation Services)",
    "cns15": "Number of jobs in NAICS sector 61 (Educational Services)",
    "cns16": "Number of jobs in NAICS sector 62 (Health Care and Social Assistance)",
    "cns17": "Number of jobs in NAICS sector 71 (Arts, Entertainment, and Recreation)",
    "cns18": "Number of jobs in NAICS sector 72 (Accommodation and Food Services)",
    "cns19": "Number of jobs in NAICS sector 81 (Other Services [except Public Administration])",
    "cns20": "Number of jobs in NAICS sector 92 (Public Administration)",
    "cr01": "Number of jobs of workers, Race: White Alone",
    "cr02": "Number of jobs of workers, Race: Black or African American Alone ",
    "cr03": "Number of jobs of workers, Race: American Indian or Alaska Native Alone",
    "cr04": "Number of jobs of workers, Race: Asian Alone",
    "cr05": "Number of jobs of workers, Race: Native Hawaiian or Other Pacific Islander Alone",
    "cr07": "Number of jobs of workers, Race: Two or More Race Groups",
    "ct01": "Number of jobs of workers, Ethnicity: Not Hispanic or Latino",
    "ct02": "Number of jobs of workers, Ethnicity: Hispanic or Latino",
    "cd01": "Number of jobs of workers, Educational Attainment: Less than high school",
    "cd02": "Number of jobs of workers, Educational Attainment: High school or equivalent, no college",
    "cd03": "Number of jobs of workers, Educational Attainment: Some college or Associate degree",
    "cd04": "Number of jobs of workers, Educational Attainment: Bachelor's degree or advanced degree",
    "cg01": "Number of jobs of workers, Sex: Male",
    "cg02": "Number of jobs of workers, Sex: Female",
    "cfa01": "Number of jobs of workers, Firm Age: 0-1 Years",
    "cfa02": "Number of jobs of workers, Firm Age: 2-3 Years",
    "cfa03": "Number of jobs of workers, Firm Age: 4-5 Years",
    "cfa04": "Number of jobs of workers, Firm Age: 6-10 Years",
    "cfa05": "Number of jobs of workers, Firm Age: 11+ Years",
    "cfs01": "Number of jobs of workers, Firm Size: 0-19 Employees",
    "cfs02": "Number of jobs of workers, Firm Size: 20-49 Employees",
    "cfs03": "Number of jobs of workers, Firm Size: 50-249 Employees",
    "cfs04": "Number of jobs of workers, Firm Size: 250-499 Employees",
    "cfs05": "Number of jobs of workers, Firm Size: 500+ Employees"
}



import json
import io
import csv
import functools

import requests
import zipfile
import shapefile
import shapely.geometry
import shapely.ops
import pyproj

from .core import AreaFilter, GEO_URLS
from .variables import LODES_VARIABLES

class OnTheMap(requests.Session):
    BASE = 'https://onthemap.ces.census.gov'

    JOB_TYPES = {'all': 'jt00',
                 'primary': 'jt01',
                 'private': 'jt02',
                 'private primary': 'jt03'}

    def _select_area(self, geojson_geometry):
        url = self.BASE + '/cgi-bin/select.py'

        geom = shapely.geometry.shape(geojson_geometry)
        geom = project(geom)

        data = {"wkt": geom.wkt}

        response = self.post(url, json=data)

        selection_id = response.json()[0]['selection_id']

        return selection_id

    def _report(self, selection_id, area, origin, job_type, year):
        url = self.BASE + '/cgi-bin/report.py'

        if area:
            analysis_type = 'area_profile_report'
        else:
            analysis_type = 'distance_direction_report'

        report_settings = {"origin_type": origin,
                           "origin_dir": "home",
                           "analysis_type": analysis_type,
                           "ap_segment": "s000",
                           "comparison_geom": "us_plc",
                           "ac_segment": "s000",
                           "destination_rollup": "us_plc",
                           "year": [year],
                           "job_type": self.JOB_TYPES[job_type],
                           "geom_operation": "ignore",
                           "selection": [selection_id],
                           "color":["#0000AA"]}

        data = {'mode': 'generate',
                'settings': json.dumps(report_settings)}

        response = self.post(url, data=data)

        report_id = response.json()['report_id']

        return report_id

    def _area(self, report_id, year):

        url = self.BASE + '/cgi-bin/report.py'

        analysis = {"analysis_type": "area_profile_report",
                    "job_type": 'jt00',
                    "ap_segment": "s000",
                    "origin": 'home',
                    "color": ["#0000AA"]}

        params = {'report_id': report_id,
                  'settings': json.dumps(analysis),
                  'mode': 'export_geography',
                  'format': 'shp'}

        response = self.get(url, params=params)

        z = zipfile.ZipFile(io.BytesIO(response.content))

        with z.open('points_{year}.dbf'.format(year=year)) as dbf:
            reader = shapefile.Reader(dbf=io.BytesIO(dbf.read()))
            fields = [field[0] for field in reader.fields[1:]]
            for row in reader.records():
                yield dict(zip(fields, row))

    def _od(self, report_id, year):

        url = self.BASE + '/cgi-bin/report.py'

        analysis = {"labor": "s000",
                    "year": 2014,
                    "analysis_type": "distance_direction_report",
                    "job_type": 'jt01',
                    "origin_type": 'work',
                    "distance":"all",
                    "color":["#0000AA"]}

        params = {'report_id': report_id,
                  'settings': json.dumps(analysis),
                  'mode': 'export_geography',
                  'format': 'shp'}

        response = self.get(url, params=params)

        z = zipfile.ZipFile(io.BytesIO(response.content))

        with z.open('points_{year}.dbf'.format(year=year)) as dbf:
            reader = shapefile.Reader(dbf=io.BytesIO(dbf.read()))
            fields = [field[0] for field in reader.fields[1:]]
            for row in reader.records():
                yield dict(zip(fields, row))

    def _geojson(self, geojson_geometry, reader):
        blocks = AreaFilter(geojson_geometry,
                            GEO_URLS['blocks'][2010])

        data = {}
        for block in reader:
            block_id = block.pop('id')
            data[block_id] = block

        features = []
        for feature in blocks:
            block_id = feature['properties']['GEOID']
            if block_id in data:
                legible = {LODES_VARIABLES[k]: v
                           for k, v
                           in data[block_id].items()}
                feature['properties'].update(legible)
                features.append(feature)


        return {'type': "FeatureCollection", 'features': features}

        

    def area_query(self, geojson, area='work', job_type='all', year=2014):
        selection_id = self._select_area(geojson)
        report_id = self._report(selection_id, True, area, job_type, year)
        reader = self._area(report_id, year)

        for row in reader:
            yield row

    def od_query(self, geojson, origin='work', job_type='all', year=2014):
        selection_id = self._select_area(geojson)
        report_id = self._report(selection_id, False, origin, job_type, year)
        reader = self._od(report_id, year)

        for row in reader:
            yield row
            
    def residents(self, geojson, job_type='all', year=2014, return_geometry=True):
        reader = self.area_query(geojson, 'home', job_type, year)
        if return_geometry:
            return self._geojson(geojson, reader)
        else:
            return list(reader)

    def workforce(self, geojson, job_type='all', year=2014, return_geometry=True):
        reader = self.area_query(geojson, 'work', job_type, year)
        if return_geometry:
            return self._geojson(geojson, reader)
        else:
            return list(reader)

    def commutes_to(self, geojson, job_type='all', year=2014, return_geometry=True):
        reader = self.od_query(geojson, 'work', job_type, year)
        return list(reader)

    def commutes_from(self, geojson, job_type='all', year=2014, return_geometry=True):
        reader = yield from self.od_query(geojson, 'home', job_type, year)
        return list(reader)

        
def project(geom):
    return shapely.ops.transform(functools.partial(pyproj.transform,
                                                   pyproj.Proj("+init=EPSG:4326"),
                                                   pyproj.Proj("+proj=aea +lat_1=29.5 +lat_2=45.5 +lat_0=37.5 +lon_0=-96 +x_0=0 +y_0=0 +datum=NAD83 +units=m +no_defs")),
                                 geom)





import collections
from functools import lru_cache
import logging
from logging import NullHandler

import census
from census.core import supported_years

import shapely.geometry
import shapely.geos
import esridump

from .variables import GEO_URLS


logging.getLogger(__name__).addHandler(NullHandler())


class AreaFilter(object):
    def __init__(self, geojson_geometry, sub_geography_url):
        self.geo = shapely.geometry.shape(geojson_geometry)

        geo_query_args = {'geometry': ','.join(str(x) for x in self.geo.bounds),
                          'geometryType': 'esriGeometryEnvelope',
                          'spatialRel': 'esriSpatialRelEnvelopeIntersects',
                          'inSR': '4326',
                          'geometryPrecision': 9,
                          'orderByFields': 'STATE,COUNTY,TRACT,OID'}
        self.area_dumper = esridump.EsriDumper(sub_geography_url,
                                               extra_query_args=geo_query_args)

    def __iter__(self):
        for area in self.area_dumper:
            area_geo = shapely.geometry.shape(area['geometry'])
            if self.geo.intersects(area_geo):
                try:
                    intersection = self.geo.intersection(area_geo)
                except shapely.geos.TopologicalError:
                    intersection = self.geo.buffer(0).intersection(area_geo.buffer(0))
                intersection_proportion = intersection.area / area_geo.area
                if intersection_proportion > 0.01:
                    yield area, intersection_proportion


class HashDict(dict):

    def __hash__(self):
        return hash(tuple(self.items()))


class GeoClient(census.core.Client):

    @lru_cache
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)

    def state_place_tract(self, *args, **kwargs):
        '''
        Retrieve variable values for tracts in the specified state and
        incorporated place.

        Arguments:

        * fields (iterable) - Variables to retrieve
        * state (int or str) - state FIPS code (see
          https://www.census.gov/library/reference/code-lists/ansi.html#state)
        * place (int or str) - place FIPS code (see
          https://www.census.gov/library/reference/code-lists/ansi.html#place)
        * year (int) - data year
        * return_geometry (bool) - set True to return results as GeoJSON like
          object (default: False)

        Returns:

        List with dictionary for each tract containing variable values.
        '''
        return self._state_place_area(self.geo_tract, *args, **kwargs)

    def state_place_blockgroup(self, *args, **kwargs):
        '''
        Retrieve variable values for block groups in the specified state and
        incorporated place.

        Arguments:

        * fields (iterable) - Variables to retrieve
        * state (int or str) - state FIPS code (see
          https://www.census.gov/library/reference/code-lists/ansi.html#state)
        * place (int or str) - place FIPS code (see
          https://www.census.gov/library/reference/code-lists/ansi.html#place)
        * year (int) - data year
        * return_geometry (bool) - set True to return results as GeoJSON like
          object (default: False)

        Returns:

        List with dictionary for each block group containing variable values.
        '''
        return self._state_place_area(self.geo_blockgroup, *args, **kwargs)

    @supported_years(2023, 2022, 2021, 2020, 2019, 2018, 2017, 2016, 2015, 2014, 2013, 2012, 2011, 2010, 2000)
    def geo_tract(self, fields, geojson_geometry, year=None, **kwargs):
        '''
        Retrieve variable values for tracts intersecting with an arbitrary
        geometry.

        Arguments:

        * fields (iterable) - Variables to retrieve
        * geojson_geometry (dict) - Geometry object in ESPG:4326
        * year (int) - data year

        Returns:

        Generator with three-tuple for each tract containing: tract geometry,
        dictionary containing variable values, and proportion of tract that
        overlaps with the arbitrary geometry
        '''
        if year is None:
            year = self.default_year

        filtered_tracts = AreaFilter(geojson_geometry,
                                     GEO_URLS['tracts'][year])

        for tract, intersection_proportion in filtered_tracts:
            context = {'state': tract['properties']['STATE'],
                       'county': tract['properties']['COUNTY']}
            within = 'state:{state} county:{county}'.format(**context)

            tract_id = tract['properties']['TRACT']
            result = self.get(fields,
                              HashDict({'for': 'tract:{}'.format(tract_id),
                                        'in': within}),
                              year,
                              **kwargs)

            if result:
                result, = result
            else:
                result = {}

            yield tract, result, intersection_proportion

    @supported_years(2023, 2022, 2021, 2020, 2019, 2018, 2017, 2016, 2015, 2014, 2013, 2012, 2011, 2010)
    def geo_blockgroup(self, fields, geojson_geometry, year=None, **kwargs):
        '''
        Retrieve variable values for block groups intersecting with an arbitrary
        geometry.

        Arguments:

        * fields (iterable) - Variables to retrieve
        * geojson_geometry (dict) - Geometry object in ESPG:4326
        * year (int) - data year

        Returns:

        Generator with three-tuple for each block group containing: block group
        geometry, dictionary containing variable values, and proportion of block
        group that overlaps with the arbitrary geometry
        '''
        if year is None:
            year = self.default_year

        filtered_block_groups = AreaFilter(geojson_geometry,
                                           GEO_URLS['block groups'][year])

        for block_group, intersection_proportion in filtered_block_groups:
            context = {'state': block_group['properties']['STATE'],
                       'county': block_group['properties']['COUNTY'],
                       'tract': block_group['properties']['TRACT']}
            within = 'state:{state} county:{county} tract:{tract}'.format(**context)

            tract_blockgroups = self.get(fields,
                                         HashDict({'for': 'block group:*',
                                                   'in':  within}),
                                         year,
                                         **kwargs)

            result = [result for result in tract_blockgroups
                      if result['block group'] == block_group['properties']['BLKGRP']]

            if result:
                result, = result
            else:
                result = {}

            yield block_group, result, intersection_proportion

    def _state_place_area(self, method, fields, state, place, year=None, return_geometry=False, **kwargs):
        if year is None:
            year = self.default_year

        search_query = "PLACE='{}' AND STATE='{}'".format(place, state)
        place_dumper = esridump.EsriDumper(GEO_URLS['incorporated places'][year],
                                           extra_query_args={'where': search_query,
                                                             'orderByFields': 'OID'})

        try:
            place = next(iter(place_dumper))
        except TypeError as e:
            if "'<' not supported between instances of 'NoneType' and 'NoneType'" in str(e):
                raise ValueError(f'Could not find specified place "{place}" in state "{state}"')

            raise e

        logging.info(place['properties']['NAME'])
        place_geojson = place['geometry']

        areas = method(fields, place_geojson, year=year, **kwargs)

        features = []
        for i, (feature, result, _) in enumerate(areas):
            if return_geometry:
                feature['properties'].update(result)
                features.append(feature)
            else:
                features.append(result)
            if i % 100 == 0:
                logging.info('{} features'.format(i))

        if return_geometry:
            return {'type': "FeatureCollection", 'features': features}
        else:
            return features

    def geo(self, fields, geojson_geometry, year=None, resolution='tract', ignore_missing = False, **kwargs):
        if year is None:
            year = self.default_year

        fields = census.core.list_or_str(fields)

        as_acs = kwargs.get('as_acs', False)
        if as_acs:
            acs_fields = self._cross(fields)
            for field in acs_fields:
                if self._field_type(field, year) is not int:
                    raise ValueError('{} is not a variable that can be aggregated your geography'.format(field))
        else:
            for field in fields:
                if self._field_type(field, year) is not int:
                    raise ValueError('{} is not a variable that can be aggregated your geography'.format(field))

        resolutions = {'tract': self.geo_tract,
                       'blockgroup': self.geo_blockgroup}

        try:
            geo_units = resolutions[resolution]
        except KeyError:
            raise ValueError('{} is not a valid resolution. Choose one of {}'.format(resolution, resolution.keys()))

        features = geo_units(fields, geojson_geometry, year=year, **kwargs)

        return self._aggregate(fields, features, year, ignore_missing)

    def _aggregate(self, fields, features, year, ignore_missing):
        aggregated_features = collections.defaultdict(list)
        for _, feature in features:
            if ignore_missing and None in feature.values():
                continue
            for field in fields:
                aggregated_features[field].append(feature[field])

        collapsed = {}

        for field in fields:
            if field.endswith('E'):
                agg = sum(aggregated_features[field])
            elif field.endswith('M'):
                e_field = field[:-1] + 'E'
                agg = census.math.moe_of_sum(aggregated_features[e_field],
                                             aggregated_features[field])
            else:
                raise ValueError("Don't know how to aggregate this variable {}".format(feature))

            collapsed[field] = agg

        return collapsed


class GeoBlockClient(GeoClient):

    def state_place_block(self, *args, **kwargs):
        '''
        Retrieve variable values for blocks in the specified state and
        incorporated place.

        Arguments:

        * fields (iterable) - Variables to retrieve
        * state (int or str) - state FIPS code (see
          https://www.census.gov/library/reference/code-lists/ansi.html#state)
        * place (int or str) - place FIPS code (see
          https://www.census.gov/library/reference/code-lists/ansi.html#place)
        * year (int) - data year
        * return_geometry (bool) - set True to return results as GeoJSON like
          object (default: False)

        Returns:

        List with dictionary for each block containing variable values.
        '''
        return self._state_place_area(self.geo_block, *args, **kwargs)

    @supported_years(2020, 2010)
    def geo_block(self, fields, geojson_geometry, year):
        '''
        Retrieve variable values for blocks intersecting with an arbitrary
        geometry.

        Arguments:

        * fields (iterable) - Variables to retrieve
        * geojson_geometry (dict) - Geometry object in ESPG:4326
        * year (int) - data year

        Returns:

        Generator with three-tuple for each block containing: block geometry,
        dictionary containing variable values, and proportion of block that
        overlaps with the arbitrary geometry
        '''
        if year is None:
            year = self.default_year

        filtered_blocks = AreaFilter(geojson_geometry,
                                     GEO_URLS['blocks'][year])

        for block, intersection_proportion in filtered_blocks:
            context = {'state': block['properties']['STATE'],
                       'county': block['properties']['COUNTY'],
                       'tract': block['properties']['TRACT']}
            within = 'state:{state} county:{county} tract:{tract}'.format(**context)

            tract_blocks = self.get(fields,
                                    HashDict({'for': 'block:*',
                                              'in':  within}),
                                    year)

            result = [result for result in tract_blocks
                      if result['block'] == block['properties']['BLOCK']]

            if result:
                result, = result
            else:
                breakpoint()
                result = {}

            yield block, result, intersection_proportion


class ACS5Client(census.core.ACS5Client, GeoClient):
    '''
    Client to access American Community Survey 5-Year Estimates (see:
    https://www.census.gov/data/developers/data-sets/acs-5year.html)
    '''
    @supported_years(2023, 2022, 2021, 2020, 2019, 2018, 2017, 2016, 2015, 2014, 2013, 2012, 2011, 2010)
    def state_place_tract(self, *args, **kwargs):
        return super().state_place_tract(*args, **kwargs)

    @supported_years(2023, 2022, 2021, 2020, 2019, 2018, 2017, 2016, 2015, 2014, 2013, 2012, 2011, 2010)
    def state_place_blockgroup(self, *args, **kwargs):
        return super().state_place_blockgroup(*args, **kwargs)


class SF1Client(census.core.SF1Client, GeoBlockClient):
    '''
    Client to access Census Summary File data (see
    https://www.census.gov/data/datasets/2010/dec/summary-file-1.html)
    '''
    @supported_years(2010)
    def state_place_tract(self, *args, **kwargs):
        return super().state_place_tract(*args, **kwargs)

    @supported_years(2010)
    def state_place_blockgroup(self, *args, **kwargs):
        return super().state_place_blockgroup(*args, **kwargs)

    @supported_years(2010)
    def state_place_block(self, *args, **kwargs):
        return super().state_place_block(*args, **kwargs)


class PLClient(census.core.PLClient, GeoBlockClient):
    '''
    Client to access Redistricting data (see
    https://www.census.gov/programs-surveys/decennial-census/about/rdo/summary-files.2020.html)
    '''
    @supported_years(2020, 2010)
    def state_place_tract(self, *args, **kwargs):
        return super().state_place_tract(*args, **kwargs)

    @supported_years(2020, 2010)
    def state_place_blockgroup(self, *args, **kwargs):
        return super().state_place_blockgroup(*args, **kwargs)

    @supported_years(2020, 2010)
    def state_place_block(self, *args, **kwargs):
        return super().state_place_block(*args, **kwargs)



#!/usr/bin/python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError :
    raise ImportError("setuptools module required, please go to https://pypi.python.org/pypi/setuptools and follow the instructions for installing setuptools")

long_description = open('README.rst').read()

setup(
    name='census_area',
    url='https://github.com/datamade/census_area',
    long_description=long_description,
    version='0.4.6',
    author='Forest Gregg',
    author_email='fgregg@datamade.us',
    description='Census data for arbitrary geographies',
    packages=['census_area'],
    install_requires=['esridump',
                      'census',
                      'shapely',
                      'pyshp',
                      'pyproj'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Information Analysis'],
)
